﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Security.Cryptography;
using System.Text;
using System.Web;

namespace DotNet.Integrations
{
    public class ApiSecurity
    {
        private static readonly char[] HEX_CHARS = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };
        protected static readonly Int64 DEFAULT_TIMEOUT = 300 * 1000;
        private readonly IApiSettingsFactory _settingsFactory;
        private readonly String _additionalPrivateKey;
        private String _realAuth;
        private String _realSignature;

        public ApiSecurity(IApiSettingsFactory settingsFactory)
            : this(settingsFactory, null)
        {
        }

        public ApiSecurity(IApiSettingsFactory settingsFactory, String additionalPrivateKey)
        {
            this._settingsFactory = settingsFactory;
            this._additionalPrivateKey = additionalPrivateKey;
        }

        public IApiSettings VerifyRequest(Int64 accountId, string url, HttpRequest request)
        {
            try
            {
                string authFromUser = request.Params.Get("auth");
                string signatureFromUser = request.Params.Get("signature");
                NameValueCollection requestParams = request.Params;
                string requestUrl = ForceSsl(url);
                return VerifyRequest(accountId, authFromUser, signatureFromUser, requestUrl, requestParams);
            }
            catch (AuthSecurityException ase)
            {
                throw new AuthSecurityException(ase);
            }
            catch (Exception ex)
            {
                throw new Exception(ex.Message, ex.InnerException);
            }
        }

        public IApiSettings VerifyRequest(Int64? accountId, String authFromUser, String signatureFromUser, String requestUrl, NameValueCollection requestParams)
        {
            if ((!accountId.HasValue || accountId == 0) || string.IsNullOrEmpty(authFromUser))
            {
                throw new AuthSecurityException("Not authorized");
            }

            try
            {
                var authToken = new AuthToken(authFromUser);

                if (!accountId.Equals(authToken.GetAccountId()))
                {
                    throw new AuthSecurityException("AccountID does not match");
                }

                Int64? integrationId = authToken.GetIntegrationId();

                IApiSettings apiSettings = _settingsFactory.GetIntegrationSettings(accountId, integrationId);

                if (apiSettings == null || !apiSettings.IsEnabled())
                {
                    throw new AuthSecurityException("No integration settings");
                }

                Int64? timestamp = authToken.GetTimestamp();
                CheckTimeout(timestamp, apiSettings);

                _realAuth = authToken.HasAuthHash() ? GetApiKey(accountId, integrationId, timestamp, GetPrivateKey(apiSettings)) : null;
                _realSignature = GetApiSignature(GetPrivateKey(apiSettings), requestUrl, requestParams);

                Int64? realAccountId = apiSettings.GetAccountId();
                if (!HasValueAndEqual(_realAuth, authFromUser) && !HasValueAndEqual(_realSignature, signatureFromUser)
                        || !accountId.Equals(realAccountId)
                        || HasValueAndNotEqual(_realSignature, signatureFromUser))
                {
                    throw new AuthSecurityException("Invalid auth or signature");
                }

                return apiSettings;
            }
            catch (AuthSecurityException ase)
            {
                throw new AuthSecurityException(ase);
            }
            catch (Exception ex)
            {
                throw new Exception(ex.Message, ex.InnerException);
            }
        }

        public String GetAPIKey(long? accountId, long? integrationId)
        {
            IApiSettings apiSettings = this._settingsFactory.GetIntegrationSettings(accountId, integrationId);
            long timestamp = DateTime.Now.TimeOfDay.Milliseconds + (5 * 60 * 1000);

            return ApiSecurity.GetApiKey(accountId, apiSettings.GetSetupItemId(), timestamp, apiSettings.GetPrivateKey());
        }

        public static String GetApiKey(Int64? accountId, Int64? setupItemId, Int64? timestamp, String privateKey)
        {
            setupItemId = setupItemId.GetValueOrDefault(0); 
            timestamp = timestamp.GetValueOrDefault(0);
            accountId = accountId.GetValueOrDefault(0);

            String key = string.Concat(accountId, ":", setupItemId, ":", timestamp);
            if (privateKey == null) return key;

            String hash = GetEncrypted(key, privateKey);
            return string.Concat(key, ":", hash);
        }

        public static String GetApiSignature(String privateKey, String requestUrl, NameValueCollection requestParams)
        {
            String unencryptedSignature = GetUnencryptedSignature(requestUrl, requestParams);
            return GetEncrypted(unencryptedSignature, privateKey);
        }

        protected String GetPrivateKey(IApiSettings apiSettings)
        {
            if (string.IsNullOrEmpty(this._additionalPrivateKey))
            {
                return apiSettings.GetPrivateKey();
            }
            else
            {
                return string.Concat(apiSettings.GetPrivateKey(), ":", this._additionalPrivateKey);
            }
        }

        public static String ForceSsl(String url)
        {
            return url.Replace("http://", "https://");
        }

        public static void CheckTimeout(Int64? timestamp, IApiSettings apiSettings)
        {
            Int64? timeout = apiSettings.GetTimeoutMilliseconds();
            if (!timeout.HasValue)
            {
                timeout = DEFAULT_TIMEOUT;
            }
            if (!timestamp.HasValue || DateTime.Now.TimeOfDay.TotalMilliseconds - timeout > timestamp)
            {
                throw new AuthSecurityException("Expired authorization");
            }
        }

        public static bool HasValueAndEqual(String realValue, String valueFromUser)
        {
            return !string.IsNullOrEmpty(realValue) && realValue.Equals(valueFromUser, StringComparison.OrdinalIgnoreCase);
        }

        public static bool HasValueAndNotEqual(String realValue, String valueFromUser)
        {
            return !string.IsNullOrEmpty(realValue) && !realValue.Equals(valueFromUser, StringComparison.OrdinalIgnoreCase);
        }

        public static String GetUnencryptedSignature(String requestUrl, NameValueCollection requestParams)
        {
            String signed = GetAsString(requestParams.Get("signed"));
            if (string.IsNullOrEmpty(signed)) return null;

            var sb = new StringBuilder();
            sb.Append(requestUrl);

            char separator = requestUrl.IndexOf('?') >= 0 ? '&' : '?';
            String[] parts = signed.Split('&');
            var setOfSignedParameters = new HashSet<String>();
            foreach (String key in parts)
            {
                setOfSignedParameters.Add(key);

                String val = GetAsString(requestParams.Get(key));
                sb.Append(separator);
                separator = '&';
                sb.Append(HttpUtility.UrlEncode(key)).Append("=").Append(val == null ? "" : HttpUtility.UrlEncode(val));
            }

            if (!HasRequiredSignatureParams(setOfSignedParameters))
            {
                return null;
            }

            return sb.ToString();
        }

        public static bool HasRequiredSignatureParams(HashSet<String> setOfSignedParameters)
        {
            return setOfSignedParameters.Contains("auth") && setOfSignedParameters.Contains("signed");
        }

        public static String GetEncrypted(String unencrypted, String privateKey)
        {
            if (unencrypted == null)
            {
                return null;
            }

            SHA512 alg = SHA512.Create();
            byte[] encrypted = alg.ComputeHash(Encoding.UTF8.GetBytes((string.Concat(unencrypted, privateKey))));
            return ToHexString(encrypted);
        }

        public static String ToHexString(byte[] bytes)
        {
            var sb = new StringBuilder();
            foreach (byte aByte in bytes)
            {
                sb.Append(HEX_CHARS[0x0F & (aByte >> 4)]);
                sb.Append(HEX_CHARS[0x0F & aByte]);
            }
            return sb.ToString();
        }

        public static String GetAsString(Object obj)
        {
            if (obj != null && obj.GetType().IsArray && obj.GetType() == typeof(string).MakeArrayType())
            {
                var o = (string[])obj;
                return o[0];
            }
            else
            {
                if (obj == null)
                {
                    return null;
                }
                else
                {
                    return obj.ToString();
                }
            }
        }

        public bool HasInvalidAuthOrSignature(String authFromUser, String signatureFromUser)
        {
            /*
            if((realAuth == null || !realAuth.equalsIgnoreCase(authFromUser)) && (realSignature == null || !realSignature.equalsIgnoreCase(signatureFromUser))) {
                return true;
            } else if(realSignature != null && !realSignature.equalsIgnoreCase(signatureFromUser)) {
                return true;
            } else {
                return false;
            }
            */

            if ((!HasValueAndEqual(_realAuth, authFromUser) && !HasValueAndEqual(_realSignature, signatureFromUser))
                    || HasValueAndNotEqual(_realSignature, signatureFromUser))
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        public void SetRealAuth(String realAuth)
        {
            this._realAuth = realAuth;
        }

        public void SetRealSignature(String realSignature)
        {
            this._realSignature = realSignature;
        }

    }
}
