﻿using System;
using System.Web;

namespace DotNet.Integrations.Filter
{
    public abstract class ApiSecurityFilter
    {
        //private static readonly int UNAUTHORIZED = 401;
        private static readonly int SERVER_ERROR = 500;

        public static readonly String AUTH_COOKIE_NAME = "apiauth";
        public static readonly String AUTH_PARAM_NAME = "auth";
        public static readonly String SIGNED_PARAM_NAME = "signed";
        public static readonly String SIGNATURE_PARAM_NAME = "signature";
        public static readonly String AUTH_ID_COOKIE_NAME = "_id";

        private static readonly int AUTH_MAX_AGE_SECONDS = 24 * 60 * 60;

        public abstract Int64 GetAccountId(HttpRequest request);
        public abstract String GetUrl(HttpRequest request);

        private IApiSettingsFactory _settingsFactory;
        protected ApiSecurityFilter(IApiSettingsFactory settingsFactory)
        {
            this._settingsFactory = settingsFactory;
        }

        public void DoFilter(HttpRequest request, HttpResponse response)
        {
            IApiSettings settings;
            String authId = request.Params.Get("authid");
            try
            {
                if (HasAuthParam(request))
                {
                    var security = new ApiSecurity(this._settingsFactory);
                    Int64 accountId = GetAccountId(request);
                    settings = security.VerifyRequest(accountId, GetUrl(request), request);
                }
                else
                {
                    var security = new ApiSecurity(this._settingsFactory, authId);
                    settings = Verifycookie(security, request);
                }
                if (settings == null)
                {
                    DeleteCooke(response);
                    response.StatusCode = SERVER_ERROR;
                    return;
                }
            }
            catch (AuthSecurityException ase)
            {
                DeleteCooke(response);
                throw new Exception(ase.Message, ase.InnerException);
            }
            catch (Exception ex)
            {
                throw new Exception(ex.Message, ex.InnerException);
            }

            authId = (string.IsNullOrEmpty(authId) ? GenerateAuthId() : authId);
            SetCookie(authId, settings, response);
            request.Params.Set("authid", authId);
        }

        private void SetCookie(string authId, IApiSettings settings, HttpResponse response)
        {
            String privateKey = string.Concat(settings.GetPrivateKey(), ":", authId);
            String apiAuth = ApiSecurity.GetApiKey(settings.GetAccountId(), settings.GetSetupItemId(), GetCookieTimeout(), privateKey);

            var cookie = new HttpCookie(AUTH_COOKIE_NAME, apiAuth);
            cookie.Expires = DateTime.Now.AddSeconds(AUTH_MAX_AGE_SECONDS);
            cookie.Secure = true;
            response.Cookies.Add(cookie);
        }

        private string GenerateAuthId()
        {
            return Guid.NewGuid().ToString();
        }

        protected long GetCookieTimeout()
        {
            return DateTime.Now.Millisecond + (AUTH_MAX_AGE_SECONDS * 1000);
        }

        private IApiSettings Verifycookie(ApiSecurity security, HttpRequest request)
        {
            String cookieValue = GetCookieValue(request, AUTH_COOKIE_NAME);
            if (string.IsNullOrEmpty(cookieValue))
            {
                throw new AuthSecurityException("No auth cookie");
            }

            var token = new AuthToken(cookieValue);
            String url;
            try
            {
                url = GetUrl(request);
            }
            catch (Exception e)
            {
                throw new AuthSecurityException("Failed to get URL", e);
            }
            url = ApiSecurity.ForceSsl(url);
            return security.VerifyRequest(token.GetAccountId(), cookieValue, null, url, request.Params);
        }

        public static string GetCookieValue(HttpRequest request, string name)
        {
            String value = null;
            HttpCookieCollection cookies = request.Cookies;
            if (cookies.Count > 0)
            {
                var c = cookies.Get(name);
                if (c != null)
                {
                    value = c.Value;
                }
            }
            return value;
        }

        protected static String Decode(String value)
        {
            //try
            //{
            return HttpUtility.UrlDecode(value);
            //}
            //catch (UnsupportedEncodingException e)    //TODO: create exception class
            //{
            //    throw new AssertionError(e);
            //}
        }

        protected static bool HasAuthParam(HttpRequest request)
        {
            String[] auth = request.Params.GetValues("auth");
            return auth != null && auth.Length > 0;
        }

        private void DeleteCooke(HttpResponse response)
        {
            var cookie = new HttpCookie(AUTH_COOKIE_NAME, null);
            cookie.Expires = DateTime.Now.AddDays(-1d);
            cookie.Secure = true;
            response.Cookies.Add(cookie);
        }
    }
}